use std::path::PathBuf;
use std::sync::Arc;

use clap::{value_parser, Arg, ArgAction, ArgMatches, Command};
use color_eyre::eyre::Context;
use color_eyre::Result;
use futures_util::StreamExt;
use sdk::{Bundle, BundleFileType};
use tokio::fs;

use crate::cmd::util::resolve_bundle_paths;

pub(crate) fn command_definition() -> Command {
    Command::new("texture-meta")
        .about(
            "Iterates over the provided bundles and lists certain meta data.
            Primarily intended to help spot patterns between dependend data fields and values.",
        )
        .arg(
            Arg::new("bundle")
                .required(true)
                .action(ArgAction::Append)
                .value_parser(value_parser!(PathBuf))
                .help(
                    "Path to the bundle(s) to read. If this points to a directory instead \
                        of a file, all files in that directory will be checked.",
                ),
        )
    // TODO: Maybe provide JSON and CSV
    // TODO: Maybe allow toggling certain fields
}

#[tracing::instrument(skip(ctx))]
async fn handle_bundle(ctx: &sdk::Context, path: &PathBuf) -> Result<()> {
    let bundle = {
        let binary = fs::read(path).await?;
        let name = Bundle::get_name_from_path(ctx, path);
        Bundle::from_binary(ctx, name, binary)?
    };

    let bundle_dir = ctx
        .game_dir
        .as_deref()
        .map(|dir| dir.join("bundle"))
        .or_else(|| path.parent().map(|p| p.to_path_buf()))
        .unwrap_or_default();

    for f in bundle.files().iter() {
        if f.file_type() != BundleFileType::Texture {
            continue;
        }

        for (i, v) in f.variants().iter().enumerate() {
            let data_file_name = v.data_file_name();

            let data_file_length = if let Some(file_name) = data_file_name {
                let path = bundle_dir.join(file_name);

                match fs::metadata(&path).await {
                    Ok(meta) => meta.len(),
                    Err(err) => {
                        return Err(err).wrap_err_with(|| {
                            format!("Failed to open data file {}", path.display())
                        })
                    }
                }
            } else {
                0
            };

            println!(
                "{},{},{},{},{:b},{},{},{:?},{},{:#010b}",
                bundle.name().display(),
                f.name(false, None),
                f.file_type().ext_name(),
                i,
                v.property(),
                v.data().len(),
                v.external(),
                data_file_name,
                data_file_length,
                v.unknown_1(),
            );
        }
    }

    Ok(())
}

#[tracing::instrument(skip_all)]
pub(crate) async fn run(ctx: sdk::Context, matches: &ArgMatches) -> Result<()> {
    let bundles = matches
        .get_many::<PathBuf>("bundle")
        .unwrap_or_default()
        .cloned();

    let paths = resolve_bundle_paths(bundles);

    let ctx = Arc::new(ctx);

    println!(
        "Bundle Name,File Name,File Type,Variant,Property,Bundle Data Length,External,Data File,Data File Length,Unknown 1"
    );

    paths
        .for_each_concurrent(10, |p| async {
            let ctx = ctx.clone();
            async move {
                if let Err(err) = handle_bundle(&ctx, &p)
                    .await
                    .wrap_err_with(|| format!("Failed to list contents of bundle {}", p.display()))
                {
                    tracing::error!("Failed to handle bundle: {}", format!("{:#}", err));
                }
            }
            .await;
        })
        .await;

    Ok(())
}
